import warnings
warnings.filterwarnings("ignore")
import os
import time
import math
import asyncio
from tqdm import tqdm
from tenacity import (
    retry,
    wait_exponential,
    stop_after_attempt,
)
from Crypto.Hash import SHA256
from Crypto.PublicKey import RSA
from Crypto.Signature import PKCS1_v1_5
from base64 import b64encode
from langchain.schema import HumanMessage
from topicgpt.walmart_llm._chat_model import WalmartAzureOpenAiChatModels
from topicgpt.walmart_llm._embed_model import WalmartAzureOpenaiEmbeddings
from topicgpt import config


def test_config():
    if config.consumer_id is None:
        raise ValueError("Please set up config.consumer_id!")
    if config.private_key_path is None:
        raise ValueError("Please set up config.private_key_path!")
    if config.mso_llm_env is None:
        raise ValueError("Please set up config.mso_llm_env!")

test_config()
os.environ['OPENAI_API_KEY'] = config.consumer_id

def get_timestamp() -> int:
    """Create timestamp
    Returns:
        int: timestamp
    """
    return int(time.time()) * 1000
 
def sign_data(private_key: str, data: str) -> bytes:
    """Create authorization signature
    Args:
        private_key (str): Path to private key
        data (str): Additional information needed to generate key in format:
            consumer_id
            timestamp
            key version
    Returns:
        bytes: authorization signature
    """
    key = open(private_key, "r").read()
    rsakey = RSA.importKey(key)
    signer = PKCS1_v1_5.new(rsakey)
    digest = SHA256.new()
    digest.update(data.encode("utf-8"))
    sign = signer.sign(digest)
    return b64encode(sign)
 
def generate_auth_sig(consumer_id: str, private_key_path: str, key_version: str = "1"):
    """_summary_
    Args:
        consumer_id (str): Service App. consumer ID
        private_key_path (str): private key path
        key_version (str) , Defaults to "1" : SOA key version
    Returns:
        tuple: epoch_time, auth_signature
    """
    epoch_time = get_timestamp()
    data = f"{consumer_id}\n{epoch_time}\n{key_version}\n"
    auth_signature = sign_data(private_key_path, data).decode()
    return epoch_time, auth_signature

def instantiate_walmart_chat_completion(model_version="gpt-35-turbo", temperature=0.):
    """The helper function that calls LLM gateway LLM
    Args:
        temperature (_type_): _description_
        max_tokens (_type_): _description_
        top_p (_type_): _description_
        presence_penalty (_type_): _description_
        frequency_penalty (_type_): _description_
        model_version (str): LLM model version
    Returns:
        _type_: _description_
    """
    epoch_ts, auth_sig = generate_auth_sig(config.consumer_id, config.private_key_path)
    consumer_params = {
        "consumer_id": config.consumer_id,
        "consumer_timestamp": str(epoch_ts),
        "consumer_key_version": "1",
        "consumer_auth_signature": auth_sig,
        "svc_env": config.mso_llm_env,
    }
    chat_client = WalmartAzureOpenAiChatModels(
        vendor="azure-openai",
        task="chat/completions",
        model_name=model_version,
        temperature=temperature,
        **consumer_params,
    )
    return chat_client

def instantiate_walmart_embedding(model_version="text-embedding-ada-002"):
    epoch_ts, auth_sig = generate_auth_sig(config.consumer_id, config.private_key_path)
    consumer_params = {
        "consumer_id": config.consumer_id,
        "consumer_timestamp": str(epoch_ts),
        "consumer_key_version": "1",
        "consumer_auth_signature": auth_sig,
        "svc_env": config.mso_llm_env,
    }
    embedding_client = WalmartAzureOpenaiEmbeddings(vendor="azure-openai", model_name=model_version, **consumer_params)
    return embedding_client


class ChatModel:

    def __init__(self, model_name="gpt-35-turbo", temperature=0., batch_size=500):
        self.temperature = temperature
        self.model = instantiate_walmart_chat_completion(model_version=model_name, temperature=temperature)
        self.batch_size = batch_size

    def completion(self, message):
        message = [[HumanMessage(content=message)]]
        response = self.model.generate(message).generations[0][0].text
        return response

    @retry(stop=stop_after_attempt(10))
    def retry_completion(self, message):
        if self.temperature == 0.:
            raise ValueError(f"To add some uncertainty, please do not set `temperature=0.`")
        try:
            message = [[HumanMessage(content=message)]]
            response = self.model.generate(message).generations[0][0].text
            return response
        except Exception as e:
            raise ValueError(f"{e}")
        
    @retry(stop=stop_after_attempt(10))
    def retry_json_completeion(self, message, keys):
        if self.temperature == 0.:
            raise ValueError(f"To add some uncertainty, please do not set `temperature=0.`")
        try:
            message = [[HumanMessage(content=message)]]
            response = self.model.generate(message).generations[0][0].text
            json_response = eval(response)
            for key in keys:
                if key not in json_response:
                    raise ValueError(f"Response doesn't include {key}")
            return json_response
        except Exception as e:
            raise ValueError(f"{e} - Response: {json_response}")
    
    def batch_completion(self, messages):
        
        async def agenerate(model, messages):
            responses = await model.agenerate(messages)
            return responses.generations

        responses = []
        num_batch = math.ceil(len(messages) / self.batch_size)
        for idx in tqdm(range(0, len(messages), self.batch_size), total=num_batch):
            batch_messages = [[HumanMessage(content=message)] for message in messages[idx: idx+self.batch_size]]
            batch_responses = asyncio.run(agenerate(self.model, batch_messages))
            responses.extend([response[0].text for response in batch_responses])
        return responses


class AdaEmbedModel:

    def __init__(self, batch_size=500, model_name="text-embedding-ada-002"):
        self.model = instantiate_walmart_embedding(model_version=model_name)
        self.batch_size = batch_size

    @retry(wait=wait_exponential(multiplier=1, min=2, max=60))
    def retry_embed_documents(self, documents):
        return self.model.embed_documents(documents)

    def embed_documents(self, documents):
        embeddings = []
        num_batch = math.ceil(len(documents) / self.batch_size)
        for idx in tqdm(range(0, len(documents), self.batch_size), total=num_batch):
            batch_documents = documents[idx: idx+self.batch_size]
            batch_embeddings = self.retry_embed_documents(batch_documents)
            embeddings.extend(batch_embeddings)
        return embeddings
    
    def embed_query(self, document):
        return self.model.embed_query(document)


import numpy as np
from umap import UMAP
from sklearn.cluster import HDBSCAN
from sklearn.metrics.pairwise import pairwise_distances
from ..base import GeneratorMixin
from ._cluster_base import Cluster
from ..feature_extraction import TopicGenerator
    

class HDBSCANTopicGenerator(GeneratorMixin):

    def __init__(self, reduced_dim, n_neighbors, min_cluster_percent, topk=5, 
                 model_name="gpt-35-turbo", temperature=0.5, verbose=True):
        self.reduced_dim = reduced_dim
        self.n_neighbors = n_neighbors
        self.min_cluster_size = min_cluster_percent
        self.topk = topk
        self.verbose = verbose
        self.nodes = 1
        self.topic_generator = TopicGenerator(model_name, temperature)

    def _dfs_build_clusters(self, root, n_neighbors, tier):
        if self.verbose:
            print("\t" * tier, "tier:", tier, "#data:", len(root.texts), "n_neighbors:", n_neighbors)

        reducer = UMAP(n_neighbors=5, n_components=self.reduced_dim, metric='cosine')
        reduced_embeddings = reducer.fit_transform(root.embeddings)

        cluster = HDBSCAN(min_cluster_size=self.min_cluster_size, n_jobs=-1)
        cluster.fit(reduced_embeddings)
        
        for label in sorted(set(cluster.labels_)):
            indices = [i for i, clabel in enumerate(cluster.labels_) if clabel == label]
            cluster_texts = [root.texts[i] for i in indices]
            cluster_embeddings = [root.embeddings[i] for i in indices]
            if self.verbose:
                print("\t" * tier, "cluster label:", label, "#data:", len(cluster_texts))

            if len(cluster_texts) <= self.min_cluster_size:
                root.children.append(Cluster(texts=cluster_texts, embeddings=cluster_embeddings))
            else:
                next_neighbors = int(n_neighbors // 2)
                if next_neighbors < 2:
                    root.children.append(Cluster(texts=cluster_texts, embeddings=cluster_embeddings))
                else:
                    sub_root = self._dfs_build_clusters(Cluster(texts=cluster_texts, embeddings=cluster_embeddings), next_neighbors, tier+1)
                    root.children.append(sub_root.children[0] if len(sub_root.children) == 1 else sub_root)
        return root.children[0] if len(root.children) == 1 else root

    def build_cluster_tree(self, texts, embeddings):
        self.nodes = 1
        self.min_cluster_size = int(self.min_cluster_size * len(texts))
        root = Cluster(texts=texts, embeddings=embeddings)
        root = self._dfs_build_clusters(root, self.n_neighbors, tier=0)
        return root
        
    def _tree_traversal(self, root):
        if root.size is None:
            root.size = len(root.texts)
        if root.percent is None:
            root.percent = 1.

        if root.centroid is None:
            root.centroid = np.mean(np.array(root.embeddings), axis=0)
        dists = pairwise_distances(np.array([root.centroid]), np.array(root.embeddings))[0]
        root.within_cluster_dist = np.mean(dists)
        root.closest = np.argsort(dists)[:self.topk]
        response = self.topic_generator.transform(root)
        root.topic = response['topic']
        root.description = response['description']

        for child in root.children:
            child.size = len(child.texts)
            child.percent = round(child.size / root.size, 3)
            self._tree_traversal(child)

    def fill_cluster_tree(self, root):
        self._tree_traversal(root)

    def plot_cluster_tree(self, root, last=True, header=''):
        elbow = "└────"
        pipe = "│  "
        tee = "├────"
        blank = "   "
        print(f"{header}{elbow if last else tee} {root.topic} - {root.description} ({root.size} | {root.percent*100:.1f}%)")
        
        child_size = len(root.children)
        self.nodes += child_size
        if child_size > 0:
            for i, c in enumerate(root.children):
                self.plot_cluster_tree(c, header=header + (blank if last else pipe), last=i == child_size - 1)

    def predict(self, texts, embeddings):
        root = self.build_cluster_tree(texts, embeddings)
        self.fill_cluster_tree(root)
        if self.verbose:
            self.plot_cluster_tree(root)
        return root


import tiktoken
from ..base import TransformerMixin
from ..walmart_llm import ChatModel


class TopicGenerator(TransformerMixin):

    def __init__(self, model_name="gpt-35-turbo", temperature=0.):
        self.model_name = model_name
        self.model = ChatModel(model_name=model_name, temperature=temperature)
        self.encoding = tiktoken.encoding_for_model(model_name=model_name)

    def _generate_topic_and_description_for_cluster(self, cluster):
        prefix = """You are a good editor. Given some requirements and sentences, you need to summarize the main topic of those sentences and give a simple description of this topic according to those requirements. 
        Here are some requirements you MUST follow:
        1. The topic should reflect the main intent of those prompts.
        2. The topic should be less than 10 words.
        3. The description should be less than 30 words.

        Here are the sentences you need to consider all:
        {user_input}
        """
        suffix = """
        The output should be in the json format:
        {"topic": <summarize the main topic of those sentence>, "description": <output a discription for this topic>}
        """
        user_input = ""
        for i, text in enumerate([cluster.texts[idx] for idx in cluster.closest], start=1):
            if (self.model_name == "gpt-35-turbo" and len(self.encoding.encode(user_input)) < 4000) or \
               (self.model_name == "gpt-4" and len(self.encoding.encode(user_input)) < 7000):
                user_input += f"sentence {i}: {str(text).strip()}\n"
        prompt = prefix.format(user_input=user_input) + suffix
        try:
            return self.model.retry_json_completeion(prompt, keys=['topic', 'description'])
        except Exception as e:
            breakpoint()
            raise ValueError(f"Exception: {e}")
         
    def transform(self, cluster):
        return self._generate_topic_and_description_for_cluster(cluster)


from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    List,
    Optional,
)
import logging
import aiohttp
import requests
from pydantic.v1 import Field, root_validator
from langchain_community.chat_models import ChatOpenAI
from langchain.schema import (
    BaseMessage,
    ChatResult,
)
from langchain_core.callbacks import (
    AsyncCallbackManagerForLLMRun,
)
logger = logging.getLogger(__name__)


class WalmartAzureOpenAiChatModels(ChatOpenAI):

    model_name: Optional[str]
    vendor: str
    task: Optional[str]
    api_version: Optional[str]
    instance_name: Optional[str]
    deployment_name: Optional[str]
    trproduct_id: Optional[str]
    consumer_id: Optional[str]
    consumer_timestamp: Optional[str]
    consumer_key_version: Optional[str]
    consumer_auth_signature: Optional[str]
    svc_env: Optional[str]
    model_kwargs: Dict[str, Any] = Field(default_factory=dict)

    class Config:
        """Configuration for this pydantic object."""

        extra = "allow"
        arbitrary_types_allowed = True
        allow_population_by_field_name = True

    @root_validator(pre=True)
    def build_model_kwargs(cls, values: Dict) -> Dict:
        """Build model_kwargs"""
        all_required_field_names = {
            'consumer_id',
            'consumer_auth_signature',
            'task',
            'deployment_name',
            'api_version',
            'consumer_timestamp',
            'trproduct_id',
            'model_kwargs',
            'model_name',
            'svc_env',
            'consumer_key_version',
            'instance_name',
            'vendor',
        }

        extra = values.get("model_kwargs", {})
        for field_name in list(values):
            if field_name not in all_required_field_names:
                if field_name in extra:
                    raise ValueError(f"Found {field_name} supplied twice.")
                extra[field_name] = values.pop(field_name)
        values["model_kwargs"] = extra
        return values
    
    @root_validator()
    def validate_environment(cls, values: Dict) -> Dict:
        """Validate that consumer details are present."""

        if not values["model_name"] and not values["instance_name"]:
            raise ValueError("Either model_name or instance_name needs to be provided.")
        if not values["consumer_id"]:
            raise ValueError("Missing mandatory parameter consumer_id.")
        if not values["svc_env"]:
            raise ValueError("Missing mandatory parameter svc_env.")

        if values["n"] < 1:
            raise ValueError("n must be at least 1.")
        if values["n"] > 1 and values["streaming"]:
            raise ValueError("n must be 1 when streaming.")
        return values

    @property
    def _llm_type(self) -> str:
        """Return type of llm."""
        return "azureopenai-chat-models"
    
    async def _agenerate(
        self,
        messages: List[BaseMessage],
        stop: Optional[List[str]] = None,
        run_manager: Optional[AsyncCallbackManagerForLLMRun] = None,
        stream: Optional[bool] = None,
        **kwargs: Any,
    ) -> ChatResult:
        message_dicts, params = self._create_message_dicts(messages, stop)
        params = {**params, **kwargs}
        response = await _acompletion_with_retry(
            self, messages=message_dicts, run_manager=run_manager, **params
        )
        return self._create_chat_result(response)
    
    def completion_with_retry(self, **kwargs: Any) -> Any:
        _model_kwargs = self.model_kwargs or {}
        payload = {}

        if self.instance_name:
            payload["instance"] = self.instance_name
        if self.model_name:
            payload["model"] = self.model_name
        if self.api_version:
            payload["api-version"] = self.api_version
        if self.deployment_name:
            payload["deployment-name"] = self.deployment_name

        payload["task"] = "chat/completions"
        # HTTP headers for authorization

        model_params = {
            "messages": kwargs["messages"],
        }
        model_params.update(_model_kwargs)
        payload["model-params"] = model_params

        headers = {
            "Content-Type": "application/json",
            "WM_CONSUMER.ID": self.consumer_id,
            "WM_SVC.NAME": "WMTLLMGATEWAY",
            "WM_SVC.ENV": self.svc_env,
        }

        if self.consumer_timestamp != "":
            headers["WM_CONSUMER.INTIMESTAMP"] = self.consumer_timestamp
        if self.consumer_key_version != "":
            headers["WM_SEC.KEY_VERSION"] = self.consumer_key_version
        if self.consumer_auth_signature != "":
            headers["WM_SEC.AUTH_SIGNATURE"] = self.consumer_auth_signature
        if self.trproduct_id != "":
            headers["WM_TR_PRODUCT.ID"] = self.trproduct_id

        gateway_endpoint = (
            f"https://wmtllmgateway.{self.svc_env}.walmart.com/wmtllmgateway/v1/openai"
        )

        def _make_rest_call(endpoint: str, headers: dict, payload: dict):
            try:
                response = requests.post(endpoint, headers=headers, json=payload, verify=False)
            except requests.exceptions.RequestException as e:
                raise ValueError(f"Error raised by inference endpoint: {e}")

            if response.status_code != 200:
                print(response.status_code)
                raise ValueError(f"Error raised by inference API: {response.content}")
            return response.json()

        # send request
        generated_text = _make_rest_call(gateway_endpoint, headers, payload)
        return generated_text
    

async def _acompletion_with_retry(llm, **kwargs: Any) -> Any:
    _model_kwargs = llm.model_kwargs or {}
    payload = {}

    if llm.instance_name:
        payload["instance"] = llm.instance_name
    if llm.model_name:
        payload["model"] = llm.model_name
    if llm.api_version:
        payload["api-version"] = llm.api_version
    if llm.deployment_name:
        payload["deployment-name"] = llm.deployment_name

    payload["task"] = "chat/completions"
    # HTTP headers for authorization

    model_params = {
        "messages": kwargs["messages"],
#         "response_format": {"type": "json_object"}
    }
    model_params.update(_model_kwargs)
    payload["model-params"] = model_params

    headers = {
        "Content-Type": "application/json",
        "WM_CONSUMER.ID": llm.consumer_id,
        "WM_SVC.NAME": "WMTLLMGATEWAY",
        "WM_SVC.ENV": llm.svc_env,
    }

    if llm.consumer_timestamp != "":
        headers["WM_CONSUMER.INTIMESTAMP"] = llm.consumer_timestamp
    if llm.consumer_key_version != "":
        headers["WM_SEC.KEY_VERSION"] = llm.consumer_key_version
    if llm.consumer_auth_signature != "":
        headers["WM_SEC.AUTH_SIGNATURE"] = llm.consumer_auth_signature
    if llm.trproduct_id != "":
        headers["WM_TR_PRODUCT.ID"] = ""  # llm.trproduct_id

    gateway_endpoint = (
        f"https://wmtllmgateway.{llm.svc_env}.walmart.com/wmtllmgateway/v1/openai"
    )

    async def _amake_rest_call(endpoint: str, headers: dict, payload: dict):
        async with aiohttp.ClientSession(
            connector=aiohttp.TCPConnector(verify_ssl=False),
            timeout=aiohttp.ClientTimeout(total=None, sock_connect=120, sock_read=120),
        ) as session:
            try:
                async with session.post(
                    endpoint, headers=headers, json=payload
                ) as response:
                    default_result = {
                        'choices': [{'finish_reason': 'stop', 'index': 0, 'message': {'content': '', 'role': 'assistant'}}], 
                        'created': 1709962264, 'id': '', 'model': '', 'object': '', 'system_fingerprint': None, 
                        'usage': {'completion_tokens': 0, 'prompt_tokens': 0, 'total_tokens': 0},
                        'error': ''
                    }
                    if response.status != 200:
                        print(f'Error raised by inference API: {await response.text()}')
                        result = default_result
                        result['error'] = f'Error raised by inference API: {await response.text()}'
                        # raise ValueError(f'Error raised by inference API: {await response.text()}')
                    else:
                        if 'application/json' in response.headers.get('Content-Type', ''):
                            result = await response.json()
                        else:
                            print('Error: response is not in JSON')
                            result = default_result
                            result['error'] = f'Error raised by parsing response to JSON'
                            # raise ValueError(f'Error raised by parsing response to JSON')
                    return result
            except aiohttp.ClientError as e:
                raise ValueError(f"Error raised by inference endpoint: {e}")

    # send request
    generated_texts = _amake_rest_call(gateway_endpoint, headers, payload)
    return await generated_texts


import re
import sys
import math
import asyncio
from tqdm import tqdm
from azure.ai.textanalytics.aio import TextAnalyticsClient
from azure.ai.textanalytics import PiiEntityCategory
from azure.core.credentials import AzureKeyCredential
from .. import config
from ..base import TransformerMixin


class MinMaxLengthFilter(TransformerMixin):

    def __init__(self, words_range=(0, -1)):
        self.words_range = words_range
        
    def transform(self, texts, labels=None):
        min_words = self.words_range[0]
        max_words = self.words_range[1] if self.words_range[1] != -1 else sys.maxsize
        texts_tr, labels_tr = [], []
        for idx in range(len(texts)):
            words = str(texts[idx]).strip().split()
            if min_words <= len(words) <= max_words:
                texts_tr.append(texts[idx])
                if labels:
                    labels_tr.append(labels[idx])
        if labels:
            return texts_tr, labels_tr
        else:
            return texts_tr
        

class NameFilter(TransformerMixin):

    batch_size = 5

    def __init__(self):
        self._test_config()

    def _test_config(self):
        if config.azure_key is None:
            raise ValueError("Please set up `config.azure_key`!")
        if config.azure_endpoint is None:
            raise ValueError("Please set up `config.azure_endpoint`!")

    async def _name_filter(self, texts):
        client = TextAnalyticsClient(
            endpoint=config.azure_endpoint,
            credential=AzureKeyCredential(config.azure_key),
        )
        async with client:
            response = await client.recognize_pii_entities(documents=texts, language="en", categories_filter = [PiiEntityCategory.PERSON], verify_ssl=False)
            filtered_texts = [re.sub(r'[*]+', 'PersonName', doc.redacted_text) for doc in response if not doc.is_error]
        await client.close()
        return filtered_texts

    def transform(self, texts):
        texts_tr = []
        num_batch = math.ceil(len(texts) / self.batch_size)
        for idx in tqdm(range(0, len(texts), self.batch_size), total=num_batch):
            texts_tr.extend(asyncio.run(self._name_filter(texts[idx: idx+self.batch_size])))
        return texts_tr


from setuptools import setup, find_packages

with open("requirements.txt", "r") as f:
    install_requires = f.read().splitlines()

setup(
    name='wm_topicgpt',
    version='0.0.5',
    description='This is a package to generate topics for the text corpus.',
    packages=find_packages(),
    install_requires=install_requires,
    python_requires='>=3.9',
)


from __future__ import annotations

import logging
import os
import requests
from requests.exceptions import SSLError
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Literal,
    Optional,
    Sequence,
    Set,
    Tuple,
    Union,
    Mapping
)

import numpy as np
from pydantic import BaseModel, model_validator, Field
from langchain.embeddings.base import Embeddings
from langchain_community.embeddings import openai
from langchain_community.embeddings import OpenAIEmbeddings
import tiktoken

logger = logging.getLogger(__name__)

def _embed_with_retry(embeddings: WalmartAzureOpenaiEmbeddings, **kwargs) -> Any:
    payload = {}

    if embeddings.instance_name:
        payload["instance"] = embeddings.instance_name        
    if embeddings.model_name:
        payload["model"] = embeddings.model_name
    if embeddings.api_version:
        payload["api-version"] = embeddings.api_version
    if embeddings.deployment_name:
        payload["deployment-name"] = embeddings.deployment_name
    # HTTP headers for authorization

    payload["task"] = "embeddings"
    model_params = {
        "input": kwargs["input"],
    }
    payload["model-params"] = model_params

    headers = {
        "Content-Type": "application/json",
        "WM_CONSUMER.ID": embeddings.consumer_id,
        "WM_SVC.NAME": "WMTLLMGATEWAY",
        "WM_SVC.ENV": embeddings.svc_env,
    }
    if embeddings.consumer_timestamp != "":
        headers["WM_CONSUMER.INTIMESTAMP"] = embeddings.consumer_timestamp
    if embeddings.consumer_key_version != "":
        headers["WM_SEC.KEY_VERSION"] = embeddings.consumer_key_version
    if embeddings.consumer_auth_signature != "":
        headers["WM_SEC.AUTH_SIGNATURE"] = embeddings.consumer_auth_signature
    if embeddings.trproduct_id != "":
        headers["WM_TR_PRODUCT.ID"] = embeddings.trproduct_id

    endpoint = f"https://wmtllmgateway.{embeddings.svc_env}.walmart.com/wmtllmgateway/v1/openai"

    try:
        try:
            if "REQUEST_CA_BUNDLE" in os.environ:
                response = requests.post(
                    endpoint, headers=headers, json=payload, verify=os.environ.get("REQUEST_CA_BUNDLE")
                )
            else:
                response = requests.post(
                    endpoint, headers=headers, json=payload, verify=True
                )
        except SSLError:
            # print("SSLError. Set REQUEST_CA_BUNDLE to use secure ssl. Falling back to insecure api call.")
            response = requests.post(
                    endpoint, headers=headers, json=payload, verify=False
                )
    except requests.exceptions.RequestException as e:
        print(f"Error raised by inference endpoint: {e}")
        raise ValueError(f"Error raised by inference endpoint: {e}")

    if response.status_code != 200:
        print(f"Error raised by inference API: {response.content}")
        raise ValueError(
            f"Error raised by inference API: {response.content}"
        )
    return response.json()

## override default behaviour
openai.embed_with_retry = _embed_with_retry

class WalmartEmbeddings(BaseModel, Embeddings):

    model_name: Optional[str]
    vendor: str
    task: Optional[str]
    api_version: Optional[str]
    instance_name: Optional[str]
    deployment_name: Optional[str]
    model_kwargs: Dict[str, Any] = Field(default_factory=dict)

    class Config:
        """Configuration for this pydantic object."""
        extra = 'allow'
        arbitrary_types_allowed = True
        populate_by_name = True
    
    def __new__(cls, **data: Any):  # type: ignore
        """Initialize the OpenAI object."""
        vendor = data.get("vendor", "")
        if vendor == "azure-openai":
            return WalmartAzureOpenaiEmbeddings(**data)
        return super().__new__(cls)

    @model_validator(mode="before")
    def build_model_kwargs(cls, values: Dict) -> Dict:
        """Build model_kwargs"""
        all_required_field_names = {field.alias for field in cls.__fields__.values()}

        extra = values.get("model_kwargs", {})
        for field_name in list(values):
            if field_name not in all_required_field_names:
                if field_name in extra:
                    raise ValueError(f"Found {field_name} supplied twice.")
                extra[field_name] = values.pop(field_name)
        values["model_kwargs"] = extra
        return values

    @model_validator(mode="before")
    def validate_environment(cls, values: Dict) -> Dict:
        """Validate that model api_token and endpoint exists in the environment."""
        _vendor = values["vendor"]
        if _vendor not in ["wmt", "azure-openai"]:
            raise ValueError(f"Unsupported vendor {_vendor}.")

        if not values["model_name"]:
            raise ValueError(f"With Walmart (wmt) as vendor, model_name parameter is mandatory.")
        if not values["consumer_id"]:
            raise ValueError("Missing mandatory parameter consumer_id.")
        if not values["svc_env"]:
            raise ValueError("Missing mandatory parameter svc_env.")
        
        return values

    @property
    def _identifying_params(self) -> Mapping[str, Any]:
        """Get the identifying parameters."""
        return {
            **{
                "model_name": self.model_name,
                "vendor": self.vendor,
                "task": self.task,
                "api_version": self.api_version,
                "instance_name": self.instance_name,
                "deployment_name": self.deployment_name,
            },
            **{"model_kwargs": self.model_kwargs}
        }

    @property
    def _llm_type(self) -> str:
        """Return type of llm."""
        return "walmart-embeddings"
    
    def embed_documents(
        self, texts: List[str], chunk_size: Optional[int] = 0
    ) -> List[List[float]]:
        """Call out to OpenAI's embedding endpoint for embedding search docs.
        Args:
            texts: The list of texts to embed.
            chunk_size: The chunk size of embeddings. If None, will use the chunk size
                specified by the class.
        Returns:
            List of embeddings, one for each text.
        """
        raise NotImplementedError("Embeddings for wmt vendor are not supported.")

    def embed_query(self, text: str) -> List[float]:
        """Call out to OpenAI's embedding endpoint for embedding query text.
        Args:
            text: The text to embed.
        Returns:
            Embedding for the text.
        """
        raise NotImplementedError("Embeddings for wmt vendor are not supported.")
    

class WalmartAzureOpenaiEmbeddings(OpenAIEmbeddings):

    model_name: str = Field(alias="model")
    vendor: str
    api_version: Optional[str]
    instance_name: Optional[str]
    deployment_name: Optional[str]
    consumer_id: Optional[str]
    consumer_timestamp: Optional[str]
    consumer_key_version: Optional[str]
    consumer_auth_signature: Optional[str]
    svc_env: Optional[str]
    trproduct_id: Optional[str]
    model_kwargs: Dict[str, Any] = Field(default_factory=dict)

    class Config:
        """Configuration for this pydantic object."""
        extra = 'allow'
        arbitrary_types_allowed = True
        populate_by_name = True

    def validate_environment(cls, values: Dict) -> Dict:
        """Validate parameters"""
        if not values["model_name"]:
            raise ValueError("Missing mandatory parameter model_name.")
        if not values["consumer_id"]:
            raise ValueError("Missing mandatory parameter consumer_id.")
        if not values["svc_env"]:
            raise ValueError("Missing mandatory parameter svc_env.")
        return values

    @property
    def _invocation_params(self) -> Dict:
        azureopenai_args = {
            "model_name": self.model_name,
            "headers": self.headers,
            "instance_name": self.instance_name,
            "deployment_name": self.deployment_name,
            "api_version": self.openai_api_version,
        }
        return azureopenai_args





from sklearn.feature_extraction.text import TfidfVectorizer
from ..base import TransformerMixin
from ..walmart_llm import ChatModel


class TfidifKeywordsExtractor(TransformerMixin):

    def __init__(self, ngram_range=(1, 2), topk=10):
        self.ngram_range = ngram_range
        self.topk = topk

    def transform(self, texts, labels=None, **params):
        if labels is None:
            labels = list(range(len(texts)))
        
        # build corpus
        uni_labels = list(set(labels))
        n_classes = len(uni_labels)
        corpus = [""] * n_classes
        for text, label in zip(texts, labels):
            corpus[uni_labels.index(label)] += (" " + str(text))

        vectorizer = TfidfVectorizer(ngram_range=self.ngram_range, stop_words='english')
        tfidf = vectorizer.fit_transform(corpus)

        # extract keywords
        keywords = []
        feature_names = vectorizer.get_feature_names_out()
        for idx in range(n_classes):
            tfidf_values = tfidf[idx].toarray().flatten()
            sorted_indices = tfidf_values.argsort()[::-1]
            top_keywords_indices = sorted_indices[:self.topk]
            top_keywords = [feature_names[idx] for idx in top_keywords_indices]
            keywords.append(top_keywords)
        
        return keywords, labels


class LlmKeywordsExtractor(TransformerMixin):

    def __init__(self, topk=10, model_name="gpt-35-turbo", temperature=0., batch_size=500):
        self.topk = topk
        self.template = "Extract at most {topk} keywords: {text}"
        self.model = ChatModel(model_name=model_name, temperature=temperature, batch_size=batch_size)
    
    def transform(self, texts, **params):
        prompts = [self.template.format(topk=self.topk, text=text) for text in texts]
        keywords = self.model.batch_completion(prompts)
        return keywords


import sys
sys.path.append("..")

#-------- Test Config ------------
from topicgpt import config

config.azure_key = "6ea086b8344b41338f484502e3f77cd9"
config.azure_endpoint = "https://my-assistant-text-analyzer-dev-66b3fa72-cognitive.cognitiveservices.azure.com/"

config.consumer_id = "719f6b98-cfb3-4048-9bd0-095b57b3257a"
config.private_key_path = "private_key.pem"
config.mso_llm_env = "stage"

#------- Test ChatModel ----------
# from topicgpt.walmart_llm import ChatModel
# llm = ChatModel(model_name="gpt-35-turbo", temperature=0.1, batch_size=500)

# message = "Who is American's president?"
# response = llm.retry_completion(message)

# message = "Question: Who is American's president? Answere in the json format as follows: {'Question': [put quesiont here], 'Answer': [put answer here]}"
# response = llm.retry_json_completeion(message, ['Question', 'Answer'])

# messages = ["Who is American's president?"] * 100000
# response = llm.batch_completion(messages)
# print(len(response))

#------- Test EmbedModel---------
# from topicgpt.walmart_llm import EmbedModel
# from topicgpt.embedding import BGEEmbedModel
# llm = BGEEmbedModel(batch_size=500, device="mps")
# llm = EmbedModel(model_name="text-embedding-ada-002", batch_size=500)

# message = "Who is American's president?"
# response = llm.embed_query(message)

# message = ["Who is American's president?"] * 10000
# response = llm.embed_documents(message)
# print(len(response))


# message = "Who is American's resident?"
# response = llm.embed_query(message)


#------- Test preprocessing -----------
# from topicgpt.preprocessing import MinMaxLengthFilter, NameFilter

# texts = ["I'm Alex", "You are good girl, please don't talk"] * 100

# filter = MinMaxLengthFilter(words_range=(0, -1))
# texts1 = filter.transform(texts)
# print(texts1)

# filter = NameFilter()
# texts2 = filter.transform(texts1)
# print(texts2)

#-------- Test feature_extraction ---------
# from topicgpt.feature_extraction import TfidifKeywordsExtractor, LlmKeywordsExtractor
# texts = ["I'm Alex", "You are good girl, please don't talk", "401k"] * 100

# extractor = TfidifKeywordsExtractor()
# r_text, r_label = extractor.transform(texts)
# print(len(r_text), len(r_label))

# extractor = LlmKeywordsExtractor()
# r_text = extractor.transform(texts)
# breakpoint()